(* This is code to convert an AST into a printable string. *)

structure AstPrint = struct

local open Ast in

fun topLevelToString(TopLevelDecl(decl)) = declToString(decl) ^ ";"
  | topLevelToString(TopLevelExpr(expr)) = exprToString(expr) ^ ";"

and declToString(ValDecl{lhs,rhs}) =
	"val " ^ patternToString(lhs) ^ " = " ^ exprToString(rhs)
  | declToString(FunDecl(fcases)) = "fun " ^ casesToString(fcases)

and casesToString([]) = ""
  | casesToString([fcase]) = caseToString(fcase)
  | casesToString(fcase::fcase2::fcases) =
	caseToString(fcase) ^ "\n  | " ^ casesToString(fcase2::fcases)

and caseToString{name,resType,formal,body} =
	name ^ patternToParenthesizedString(formal) ^ 
	typeOptionToString(resType) ^ " = " ^ exprToString(body)

and declsToString([]) = ""
  | declsToString([decl]) = declToString(decl)
  | declsToString(decl::decl2::decls) = 
	declToString(decl) ^ "\n\t" ^ declsToString(decl2::decls)

and exprToString(VarExpr(name)) = name
  | exprToString(IntExpr(i)) = Int.toString(i)
  | exprToString(StringExpr(s)) = "\"" ^ s ^ "\""
  | exprToString(BoolExpr(true)) = "true"
  | exprToString(BoolExpr(false)) = "false"
  | exprToString(NilExpr{typ=NONE}) = "nil"
  | exprToString(NilExpr{typ=SOME typ}) = "(nil:" ^ typeToString(typ) ^ ")"
  | exprToString(TupleExpr(exprs)) = "(" ^ exprsToString(exprs) ^ ")"
  | exprToString(UnopExpr{operator,arg}) =
	"(" ^ operator ^ " " ^ exprToString(arg) ^ ")"
  | exprToString(BinopExpr{arg1,operator,arg2}) =
	"(" ^ exprToString(arg1) ^ " " ^ operator ^ " " ^ 
	exprToString(arg2) ^ ")"
  | exprToString(CallExpr{function,arg}) = function ^ " " ^ exprToString(arg)
  | exprToString(IfExpr{testExpr,thenExpr,elseExpr}) =
	"if " ^ exprToString(testExpr) ^ " then " ^ exprToString(thenExpr) ^
	" else " ^ exprToString(elseExpr)
  | exprToString(LetExpr{decls,body}) =
	"let " ^ declsToString(decls) ^ " in " ^ exprToString(body) ^ " end"

and exprsToString([]) = ""
  | exprsToString([expr]) = exprToString(expr)
  | exprsToString(expr::expr2::exprs) =
	exprToString(expr) ^ ", " ^ exprsToString(expr2::exprs)

and patternToParenthesizedString(p as TuplePattern(_)) = patternToString(p)
  | patternToParenthesizedString(p as ConsPattern(_)) = patternToString(p)
  | patternToParenthesizedString(p as NilPattern{typ=SOME _}) =
	patternToString(p)
  | patternToParenthesizedString(p as WildcardPattern{typ=SOME _}) =
	patternToString(p)
  | patternToParenthesizedString(p) = "(" ^ patternToString(p) ^ ")"

and patternToString(VarPattern{name,typ}) = name ^ typeOptionToString(typ)
  | patternToString(WildcardPattern{typ=NONE}) = "_"
  | patternToString(WildcardPattern{typ=SOME typ}) =
	"(_:" ^ typeToString(typ) ^ ")"
  | patternToString(IntPattern(i)) = Int.toString(i)
  | patternToString(StringPattern(s)) = "\"" ^ s ^ "\""
  | patternToString(BoolPattern(true)) = "true"
  | patternToString(BoolPattern(false)) = "false"
  | patternToString(NilPattern{typ=NONE}) = "nil"
  | patternToString(NilPattern{typ=SOME typ}) =
	"(nil:" ^ typeToString(typ) ^ ")"
  | patternToString(TuplePattern(patterns)) =
	"(" ^ patternsToString(patterns) ^ ")"
  | patternToString(ConsPattern{head,tail}) =
	"(" ^ patternToString(head) ^ " :: " ^ patternToString(tail) ^ ")"

and patternsToString([]) = ""
  | patternsToString([pattern]) = patternToString(pattern)
  | patternsToString(pattern::pattern2::patterns) =
	patternToString(pattern) ^ ", " ^ patternsToString(pattern2::patterns)

and typeOptionToString(NONE) = ""
  | typeOptionToString(SOME typ) = ":" ^ typeToString(typ)

and typeToString(IntType) = "int"
  | typeToString(StrType) = "string"
  | typeToString(BoolType) = "bool"
  | typeToString(UnitType) = "unit"
  | typeToString(TypeVar(name)) = name
  | typeToString(TupleType(typeList)) = "(" ^ typesToString(typeList) ^ ")"
  | typeToString(ListType(typ)) = typeToString(typ) ^ " list"
  | typeToString(FunctionType{argType,resType}) =
	"(" ^ typeToString(argType) ^ " -> " ^ typeToString(resType) ^ ")"

and typesToString([]) = ""
  | typesToString([typ]) = typeToString(typ)
  | typesToString(typ1::typ2::types) =
	typeToString(typ1) ^ " * " ^ typesToString(typ2::types)

end

end
