(*

An implementation of environments, which how the evaluator stores the
mapping from variables to their contents.  The Assoc_List.iterate
calls are good examples of how first-class functions can be used.

*)

structure Env = struct

local
  open Value
  open ValuePrint
in

(* an environment stores a list of bindings, one for each variable
that's bound, mapping each variable to its current Value *)
type Environment = (string, Value) Assoc_List.assoc_list

(* empty_environment is the initial environment with no bound variables *)
val empty_environment:Environment = Assoc_List.empty

(* add_binding adds a name->value binding to the environment,
returning a new extended environment. *)
fun add_binding(env:Environment, name:string, value:Value):Environment =
    Assoc_List.store(env, name, value)

(* new_binding creates a new environment containing a single binding
of name->value *)
fun new_binding(name:string, value:Value):Environment =
    add_binding(empty_environment, name, value)

(* combine_envs takes two environments and adds them together to form
a new, merged environment containing all the bindings of the two
combined environments.  if any name is defined in both enviroments,
the binding in env2 takes precedence (shadows) that in env1 *)
fun combine_envs(env1:Environment, env2:Environment):Environment =
    let fun combine_one((name:string, value:Value), env):Environment =
	        add_binding(env, name, value)
    in Assoc_List.iterate(combine_one, env1, env2) end

(* NotFound is the exception raised if lookups or removes fail *)
val NotFound = Assoc_List.NotFound

(* lookup_binding takes an environment and a name, and returns the
value bound to that name in the environment, or raises NotFound if the
name was unbound *)
fun lookup_binding(env:Environment, name:string):Value =
    Assoc_List.fetch(env, name)

(* envToString converts an environment to a printable string *)
fun envToString(env:Environment):string =
    let fun oneToString((name:string, value:Value), prefix:string):string =
                prefix ^ "val " ^ name ^ " = " ^ valueToString(value) ^"\n"
    in Assoc_List.iterate(oneToString, "", env) end

end

end
