(*

This file defines the way to evaluate all the different kinds of ASTs.

It's a partial implementation; many cases of expression evaluation and
pattern-matching have been stubbed out with ImplementMe exceptions.

*)

structure Eval = struct

local
  open Ast
  open AstPrint
  open Value
  open ValuePrint
  open Env
in


(* if a case is initially unimplemented, but needs to be implemented
for successful completion of the project, then it raises the
ImplementMe exception, passing a description of the unimplemented
feature as the argument *)

exception ImplementMe of string


(* if any eval functions encounter a run-time error, then they raise
the EvalError exception, with a string explaining the error *)

exception EvalError of string


(* if pattern-matching fails, it raises the MatchFailure exception *)

exception MatchFailure


(* eval takes the top-level decl or expr, and the top-level
environment (i.e., the bindings of global variables to their values),
and then evaluates the top-level ast to produce a set of new bindings
(possibly of "it"), which is returned to the caller *)

fun eval(ast:TopLevel, env:Environment):Environment =
    evalTopLevel(ast, env)


(* evalTopLevel does the work of eval *)

and evalTopLevel(TopLevelDecl(decl), env):Environment =
        evalDecl(decl, env)
  | evalTopLevel(TopLevelExpr(expr), env) = 
        (* treat a top-level expression as if it were "val it = expr" *)
        let val it_decl = ValDecl{lhs=VarPattern{name="it", typ=NONE},
			          rhs=expr}
	in evalDecl(it_decl, env) end


(* evalDecl evaluates a declaration to produce a set of new bindings *)

and evalDecl(ValDecl{lhs:Pattern,rhs:Expr}, env):Environment =
        (* to evaluate a var decl, first evaluate its rhs expression
           to produce a result Value, then match the rhs value against
	   the lhs pattern, returning any variables that are bound by
	   the pattern-match *)
        let val value:Value = evalExpr(rhs, env)
            val bindings:Environment =
	        matchPattern(lhs, value)
		    handle MatchFailure =>
		        raise EvalError("cannot match " ^ exprToString(rhs) ^
					" against " ^ patternToString(lhs) ^
					" in val binding")
        in bindings end

  | evalDecl(FunDecl(cases), env) =
        (* to evaluate a function decl, simply construct a value
	   representing the function (along with the current
	   environment, to support lexical scoping when the function is later
	   called) and then return a binding of the function name to
	   the function value *)
	raise ImplementMe("function decls")


(* evalExpr evaluates an expression in an environment to produce a
result Value *)

and evalExpr(VarExpr(var_name:string), env):Value =
        (* to evaluate a variable reference, just look up the variable's
           binding in the current environment (raise an error if the variable
           isn't bound in the current environment) *)
        (lookup_binding(env, var_name) 
            handle NotFound => 
	        raise EvalError("variable " ^ var_name ^ " not bound"))

        (* to evaluate a constant, just return the corresponding value *)
  | evalExpr(IntExpr(v), env):Value = IntValue(v)
  | evalExpr(StringExpr(v), env):Value = StringValue(v)
  | evalExpr(BoolExpr(v), env):Value = BoolValue(v)
  | evalExpr(NilExpr{typ}, env):Value = ListValue(nil)

  | evalExpr(TupleExpr(exprs:Expr list), env):Value =
        (* to evaluate a tuple expression, evaluate each of the
           subexpressions, then construct the tuple value *)
	raise ImplementMe("tuple exprs")

  | evalExpr(UnopExpr{operator:string, arg:Expr}, env):Value =
        (* to evaluate a unary operator expression, evaluate the
	   argument, then apply the operator to the argument.  uses
	   applyUnop below. *)
	raise ImplementMe("unop exprs")

  | evalExpr(BinopExpr{arg1:Expr, operator:string, arg2:Expr}, 
             env):Value =
        (* to evaluate a binary operator expression, evaluate the
	   arguments, then apply the operator to the arguments.  uses
	   applyBinop below. *)
	raise ImplementMe("binop exprs")

  | evalExpr(CallExpr{function:string, arg:Expr}, env):Value =
        (* to evaluate a function call, first find the function value
           in the environment (like a variable reference expression),
	   and verify that it's actually a function value.  then
           evaluate the argument expression.  finally, evaluate the
	   function's cases (described below), in the environment
	   lexically enclosing the callee function's declaration. *)
	raise ImplementMe("call exprs")

  | evalExpr(IfExpr{testExpr, thenExpr, elseExpr}, env):Value =
        (* to evaluate an if, first evaluate the testExpr.  if it's
	   true, evaluate and return the thenExpr, otherwise evaluate and 
	   return the elseExpr. *)
	raise ImplementMe("if exprs")

  | evalExpr(LetExpr{decls:Decl list, body:Expr}, env):Value =
        (* to evaluate a let, first evaluate the decls to produce a
	   set of new bindings.  then extend the current environment "env" with
	   the new bindings to produce a new extended environment.  finally,
	   evaluate the body expression in the environment including all the 
	   new bindings, and return the result of the body evaluation. *)
	raise ImplementMe("let exprs")


(* matchPattern takes a pattern and a value and does pattern-matching.
if successful, it returns a list of bindings of variables in the
pattern to their corresponding pieces of the matched value.  if
unsuccessful, it raises a MatchFailure exception *)

and matchPattern(VarPattern{name:string,typ:Type option},
		 value:Value):Environment =
        (* variable patterns always succeed, and make a binding for
	   the variable to the value that it matched *)
        new_binding(name, value)

  | matchPattern(WildcardPattern{typ}, value:Value):Environment =
        (* wildcard patterns always succeed, but don't create any bindings *)
        empty_environment

        (* constant patterns match equal constant values, without
	   creating any bindings *)
  | matchPattern(IntPattern(i1), IntValue(i2)):Environment =
        if i1 = i2 then empty_environment else raise MatchFailure
  | matchPattern(StringPattern(s1), StringValue(s2)):Environment =
        if s1 = s2 then empty_environment else raise MatchFailure
  | matchPattern(BoolPattern(b1), BoolValue(b2)):Environment =
        if b1 = b2 then empty_environment else raise MatchFailure
  | matchPattern(NilPattern{typ}, ListValue(nil)):Environment =
        empty_environment

  | matchPattern(TuplePattern(patterns), TupleValue(values)) =
        (* a tuple pattern matches a tuple value if its components match *)
	raise ImplementMe("tuple patterns")

  | matchPattern(ConsPattern{head:Pattern, tail:Pattern}, 
		 ListValue(first::rest)) =
        (* a :: pattern matches a non-empty list value if its head and
	   tail patterns match the first and rest elements of the list,
	   respectively *)
	raise ImplementMe("cons patterns")

  | matchPattern(_, _) =
        (* otherwise, things don't match *)
	raise MatchFailure


(* applyUnop and applyBinop carry out the work of applying an operator
to its argument value(s), returning a result value.  they raise
EvalError if the type(s) of the argument value(s) are not appropriate
for the operator. *)

and applyUnop("~",   IntValue(arg) ):Value = IntValue (~ arg)
  | applyUnop("not", BoolValue(arg)):Value = BoolValue(not arg)
  | applyUnop(unop, v) =
        raise EvalError("mistyped argument " ^ valueToString(v) ^
			" to " ^ unop)


and applyBinop(IntValue(arg1), "+", IntValue(arg2)) = IntValue(arg1 + arg2)
  | applyBinop(IntValue(arg1), "-", IntValue(arg2)) = IntValue(arg1 - arg2)
  | applyBinop(IntValue(arg1), "*", IntValue(arg2)) = IntValue(arg1 * arg2)
  | applyBinop(IntValue(arg1), "div", IntValue(arg2)) =
        IntValue(arg1 div arg2)
  | applyBinop(IntValue(arg1), "mod", IntValue(arg2)) =
        IntValue(arg1 mod arg2)
  | applyBinop(IntValue(arg1), "<",  IntValue(arg2)) = BoolValue(arg1 <  arg2)
  | applyBinop(IntValue(arg1), "<=", IntValue(arg2)) = BoolValue(arg1 <= arg2)
  | applyBinop(IntValue(arg1), ">=", IntValue(arg2)) = BoolValue(arg1 >= arg2)
  | applyBinop(IntValue(arg1), ">",  IntValue(arg2)) = BoolValue(arg1 >  arg2)

  | applyBinop(StringValue(arg1), "^", StringValue(arg2)) =
        StringValue(arg1 ^ arg2)
  | applyBinop(StringValue(arg1), "<",  StringValue(arg2)) =
        BoolValue(arg1 <  arg2)
  | applyBinop(StringValue(arg1), "<=", StringValue(arg2)) =
        BoolValue(arg1 <= arg2)
  | applyBinop(StringValue(arg1), ">=", StringValue(arg2)) =
        BoolValue(arg1 >= arg2)
  | applyBinop(StringValue(arg1), ">",  StringValue(arg2)) =
        BoolValue(arg1 >  arg2)

  | applyBinop(arg1, "::", ListValue(arg2)) =
        ListValue(arg1 :: arg2)

  | applyBinop(v1, "=", v2) = BoolValue(is_equal(v1, "=", v2))

  | applyBinop(v1, "<>", v2) = BoolValue(not(is_equal(v1, "<>", v2)))

  | applyBinop(v1, binop, v2) =
        raise EvalError("mistyped arguments " ^ valueToString(v1) ^ " and " ^
			valueToString(v2) ^ " to " ^ binop)


(* is_equal is a helper function that implements the = operator on
pairs of values of the same type.  it raises an EvalError if the
arguments are of different types or if functions are compared. *)

and is_equal(IntValue(arg1), opname:string, IntValue(arg2)) = arg1 = arg2
  | is_equal(StringValue(arg1), opname, StringValue(arg2)) = arg1 = arg2
  | is_equal(BoolValue(arg1), opname, BoolValue(arg2)) = arg1 = arg2
  | is_equal(TupleValue(values1), opname, TupleValue(values2)) =
        are_equal(values1, opname, values2)
  | is_equal(ListValue(values1), opname, ListValue(values2)) =
        are_equal(values1, opname, values2)
  | is_equal(v1 as FunValue(arg1), opname, v2 as FunValue(arg2)) = 
        raise EvalError("cannot apply " ^ opname ^ " to function values " ^
			valueToString(v1) ^ " and " ^ valueToString(v2))
  | is_equal(v1, opname, v2) =
        raise EvalError("mistyped arguments " ^ valueToString(v1) ^ " and " ^
			valueToString(v2) ^ " to " ^ opname)


(* are_equal tests whether corresponding elements of two lists are equal. *)

and are_equal([], opname:string, []) = true
  | are_equal(first1::rest1, opname, first2::rest2) =
        is_equal(first1, opname, first2) andalso
	are_equal(rest1, opname, rest2)
  | are_equal(_, opname, _) = false


end

end
